//<copyright-disclaimer-start>
//  **************************************************************************************************************
//  * © 2025 Advanced Micro Devices, Inc. All rights reserved.                                                   *
//  * DISCLAIMER                                                                                                 *
//  * The information contained herein is for informational purposes only, and is subject to change              *
//  * without notice. While every precaution has been taken in the preparation of this document, it              *
//  * may contain technical inaccuracies, omissions and typographical errors, and AMD is under no                *
//  * obligation to update or otherwise correct this information.  Advanced Micro Devices, Inc. makes            *
//  * no representations or warranties with respect to the accuracy or completeness of the contents of           *
//  * this document, and assumes no liability of any kind, including the implied warranties of noninfringement,  *
//  * merchantability or fitness for particular purposes, with respect to the operation or use of AMD            *
//  * hardware, software or other products described herein.  No license, including implied or                   *
//  * arising by estoppel, to any intellectual property rights is granted by this document.  Terms and           *
//  * limitations applicable to the purchase or use of AMD’s products are as set forth in a signed agreement     *
//  * between the parties or in AMD's Standard Terms and Conditions of Sale. GD-18                               *
//  *                                                                                                            *
//  **************************************************************************************************************
//<copyright-disclaimer-end>


#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <fstream>

#include "xrt/experimental/xrt_aie.h"
#include "xrt/experimental/xrt_kernel.h"
#include "adf/adf_api/XRTConfig.h"

// ********************************************************************************** //
// ---------------------------------------------------------------------------------- //
//                          A 7 2 -  A P P L I C A T I O N                            //
// ---------------------------------------------------------------------------------- //
// ********************************************************************************** //

// ============================================================================
// Step 1: Include the AI Engine Graph Application and XRT and ADF Headers
// ============================================================================
#include "graph.cpp"
#include "data.h"

using namespace std;

#define XRT_SUCCESS 0
#define ALL_MESSAGES

static std::vector<char>
load_xclbin(xrtDeviceHandle device, const std::string& fnm)
{
    if (fnm.empty())
        throw std::runtime_error("No xclbin speified");

    // load bit stream
    std::ifstream stream(fnm);
    stream.seekg(0,stream.end);
    size_t size = stream.tellg();
    stream.seekg(0,stream.beg);

    std::vector<char> header(size);
    stream.read(header.data(),size);

    auto top = reinterpret_cast<const axlf*>(header.data());
    if (xrtDeviceLoadXclbin(device, top))
        throw std::runtime_error("Bitstream download failed");

    return header;
}

#define SAMPLES 256

/*******************************************************************************
 ** Main Function
 *******************************************************************************/
int main(int argc, char ** argv)
{

	cout << endl;

    // ============================================================================
    // Step 2: Check Command Line Argument
    // ============================================================================
    //    o) argv[1] XCLBIN file
    // ============================================================================

#ifdef ALL_MESSAGES
    cout << "A72-Info: ============================================================= " << endl;
    cout << "A72-Info: (Step 2) Check Command Line Arguments                      " << endl;
    cout << "A72-Info: ============================================================= " << endl;
#endif

    if (argc != 2) {
        cout << "A72-Error: Incorrect command line syntax " << endl;
        cout << "A72-Info:  Usage: " << argv[0] << " <XCLBIN_File> " << endl << endl;
        return EXIT_FAILURE;
    }

    // ------------------------------------------------------------------
    // Step 3: Load XCLBIN Binary File from a disk to memory
    // ------------------------------------------------------------------
    cout << endl;
#ifdef ALL_MESSAGES
    cout << "A72-Info: ============================================================= " << endl;
    cout << "A72-Info: (Step 3) Open XLCBIN and Create Data Mover Kernel Handles "     << endl;
    cout << "A72-Info: ============================================================= " << endl;
#endif

    //Open a device and obtain its handle
    auto dhdl = xrtDeviceOpen(0);

    //Download FPGA image (xclbin) to the device
    auto xclbin = load_xclbin(dhdl, argv[1]);

    //Point to top of xclbin data
    auto top = reinterpret_cast<const axlf*>(xclbin.data());

    int sizeIn = SAMPLES/2;
    int sizeOut = SAMPLES;


  // ============================================================================
  // Step 4: Allocate buffers, create kernels, execute the graph
  // ============================================================================
#ifdef ALL_MESSAGES
  cout << endl;
  cout << "A72-Info: ============================================================= " << endl;
  cout << "A72-Info: (Step 4) Allocate buffers for Input data and Results " << endl;
  cout << "A72-Info:          in Global Memory"              << endl;
  cout << "A72-Info: ============================================================= " << endl;
#endif

    //////////////////////////////////////////
    // input memory
    // No cache no sync seems not working. Should ask SSW team to investigate.
    //////////////////////////////////////////	

    //Allocate BOs (buffer objects) of requested size with appropriate flags
    xrtBufferHandle in_bohdl = xrtBOAlloc(dhdl, sizeIn * sizeof(int16_t) * 2, 0, 0);

    //Memory map BOs into user's address space (DDR Memory)
    auto in_bomapped = reinterpret_cast<uint32_t*>(xrtBOMap(in_bohdl));

    //Read the data from the file data.h // Defines variable  short int cint16input[]
    cout << "A72-Info: Storing input data in DDR ..." << endl;
    memcpy(in_bomapped, cint16Input, sizeIn * sizeof(int16_t) * 2);
    // printf("Input memory virtual addr 0x%llx\n", in_bomapped);

    //Synchronize buffer contents in requested direction (XCL_BO_SYNC_BO_TO_DEVICE)
    xrtBOSync(in_bohdl, XCL_BO_SYNC_BO_TO_DEVICE, sizeIn * sizeof(int16_t) * 2, 0);
    
    //////////////////////////////////////////
    // output memory
    //////////////////////////////////////////

    //Allocate BOs (buffer objects) of requested size with appropriate flags
    xrtBufferHandle out_bohdl = xrtBOAlloc(dhdl, sizeOut * sizeof(int), 0, 0);
    
    //Memory map BOs into user's address space (DDR Memory)
    auto out_bomapped = reinterpret_cast<uint32_t*>(xrtBOMap(out_bohdl));

    memset(out_bomapped, 0xABCDEF00, sizeOut * sizeof(int));
    // printf("Output memory virtual addr 0x%llx\n", out_bomapped);

    std::cout << "A72-Info: Processed BOs (input_mapped and output_mapped)\n";

    // ------------------------------------------------------------------
    // Step 5: 
    // 1. Create PL Kernel Handles that point to the correct location
    //           in XCLBIN file for each data mover kernel
    // 2. Execute the Data Mover Kernels 
    // ------------------------------------------------------------------

    cout << "A72-Info: Start mm2s_0, s2mm and polar_clip kernels ..." << endl;

    //////////////////////////////////////////
    // mm2s ip
    //////////////////////////////////////////

    xrtKernelHandle mm2s_khdl = xrtPLKernelOpen(dhdl, top->m_header.uuid, "mm2s");
    xrtRunHandle mm2s_rhdl = xrtRunOpen(mm2s_khdl);
    int rval = xrtRunSetArg(mm2s_rhdl, 0, in_bohdl);
    rval = xrtRunSetArg(mm2s_rhdl, 2, sizeIn);
    xrtRunStart(mm2s_rhdl);

    cout << "A72-Info: run mm2s" << endl;

    //////////////////////////////////////////
    // s2mm ip
    //////////////////////////////////////////

    xrtKernelHandle s2mm_khdl = xrtPLKernelOpen(dhdl, top->m_header.uuid, "s2mm");
    xrtRunHandle s2mm_rhdl = xrtRunOpen(s2mm_khdl);
    rval = xrtRunSetArg(s2mm_rhdl, 0, out_bohdl);
    rval = xrtRunSetArg(s2mm_rhdl, 2, sizeOut);
    xrtRunStart(s2mm_rhdl);

    cout << "A72-Info: run s2mm" << endl;

    //////////////////////////////////////////
    // polar clip ip
    //////////////////////////////////////////

    xrtKernelHandle polar_clip_khdl = xrtPLKernelOpen(dhdl, top->m_header.uuid, "polar_clip");
    xrtRunHandle polar_clip_rhdl = xrtRunOpen(polar_clip_khdl);
    rval = xrtRunSetArg(polar_clip_rhdl, 2, sizeOut);
    xrtRunStart(polar_clip_rhdl);

    cout << "A72-Info: run polar_clip" << endl;

    // ------------------------------------------------------------------
    // Step 6: Run the Graph
    // ------------------------------------------------------------------
    
    //////////////////////////////////////////
    // Graph execution for AIE
    //////////////////////////////////////////

    adf::registerXRT(dhdl, top->m_header.uuid);

    cout << "A72-Info: Graph init" << endl;
    clipgraph.init();

    cout << "A72-Info: Graph run" << endl;
    clipgraph.run(1);

    clipgraph.end();
    cout << "A72-Info: Graph end" << endl;

    // ------------------------------------------------------------------
    // Step 7: Wait for the kernels to complete 
    // ------------------------------------------------------------------
    
    //////////////////////////////////////////
    // wait for mm2s done
    //////////////////////////////////////////	

    auto state = xrtRunWait(mm2s_rhdl);
    cout << "A72-Info: mm2s completed with status(" << state << ")\n";
    xrtRunClose(mm2s_rhdl);
    xrtKernelClose(mm2s_khdl);


    //////////////////////////////////////////
    // wait for polar clip done
    //////////////////////////////////////////	

    state = xrtRunWait(polar_clip_rhdl);
    cout << "A72-Info: polar_clip completed with status(" << state << ")\n";
    xrtRunClose(polar_clip_rhdl);
    xrtKernelClose(polar_clip_khdl);


    //////////////////////////////////////////
    // wait for s2mm done
    //////////////////////////////////////////	

    state = xrtRunWait(s2mm_rhdl);
    cout << "A72-Info: s2mm completed with status(" << state << ")\n";
    xrtRunClose(s2mm_rhdl);
    xrtKernelClose(s2mm_khdl);

    xrtBOSync(out_bohdl, XCL_BO_SYNC_BO_FROM_DEVICE, sizeOut * sizeof(int), 0);
    
    // ------------------------------------------------------------------
    // Step 7: Compare the results
    // ------------------------------------------------------------------
#ifdef ALL_MESSAGES
  cout << endl;
  cout << "A72-Info: ============================================================= "   << endl;
  cout << "A72-Info: (Step 5) Verify Output Results " << endl;
  cout << "A72-Info: ============================================================= "   << endl;
#endif
    int errorCount = 0;
    {
        for (int i = 0; i < sizeOut; i++)
        {
            if (out_bomapped[i] != golden[i])
            {
                printf("A72-Info: Error found @ %d, %d != %d\n", i, out_bomapped[i], golden[i]);
                errorCount++;
            }
        }

        if (errorCount)
            printf("A72-Info: Test failed with %d errors\n", errorCount);
        else
            printf("A72-Info: Test passed\n");
    }


    // ------------------------------------------------------------------
    // Step 7: Clean up XRT
    // ------------------------------------------------------------------
#ifdef ALL_MESSAGES
    cout << endl;
    cout << "A72-Info: ============================================================= "   << endl;
    cout << "A72-Info: (Step 6) Release Allocated Resources" << endl;
    cout << "A72-Info: ============================================================= "   << endl;
#endif
    cout << "A72-Info: Releasing remaining XRT objects...\n";
    xrtBOFree(in_bohdl);
    xrtBOFree(out_bohdl);
    xrtDeviceClose(dhdl);

    return errorCount;
}
